function enforce_type(type_name, rhs, message)
   if type(rhs) ~= type_name then print(message) return nil end
   return true
end

function sanity_check_ingredient(index, ingredient)
   if not enforce_type("number", index, "An index is needed for error reporting when sanity checking ingredient.") then return nil end
   if not enforce_type("table", ingredient,
                       "Ingredient " .. index .. " is not a table with an amount and a name.") then return nil end
   if not enforce_type("number", ingredient.amount, "Ingredient " .. index .. " has NAN amount.") then return nil end
   if not enforce_type("string", ingredient.name, "Ingredient " .. index .. " has NAS name.") then return nil end
   return true
end

function sanity_check_recipe(index, recipe)
   if not index then print("An index is needed for error reporting when sanity checking recipe.") return nil end
   if not recipe then print("Cannot sanity check nil recipe.") return nil end
   if type(recipe) ~= "table" then
      print("Cannot sanity check non-table recipe.")
      return nil
   end
   local sane= true
   if not recipe.station or type(recipe.station) ~= "string" then
      print("Recipe " .. index .. " lacks a station field.  Use \"Anywhere\" to indicate a recipe active with no station.")
      sane= false
   else
      if #recipe < 1 then
         print("Recipe " .. index .. " lacks ingredients.  They must be in an array.")
         sane= false
      else
         for i, ing in ipairs(recipe) do
            if not sanity_check_ingredient(i, ing) then
               print("Recipe " .. index .. " failed sanity check.")
               sane= false
            end
         end
      end
   end
   return sane
end

function sanity_check_single_material(name, mat)
   if not name then print("scsm must be given name.") return nil end
   if not mat then print("scsm must be given mat.") return nil end
   if type(mat) ~= "table" then
      print("Material \"" .. name .. "\" is not a table.")
      return nil
   end
   local sane= true
   if not mat.name then
      print("Material \"" .. name .. "\" doesn't know itself.")
      sane= false
   end
   if not mat.recipes then
      print("Material \"" .. name .. "\" lacks a recipes table.  An empty recipes table indicates an uncraftable material.")
      sane= false
   elseif type(mat.recipes) ~= "table" then
      print("Material \"" .. name .. "\" recipes are not in a table.")
      sane= false
   else
      for i, v in ipairs(mat.recipes) do
         if not sanity_check_recipe(i, v) then
            print("Material \"" .. name .. "\" failed sanity check in recipe list.")
            sane= false
         end
      end
   end
   return sane
end

function sanity_check_materials(mats)
   if not mats then print("Cannot sanity check nil table.") return nil end
   if type(mats) ~= "table" then
      print("Materials must be a table.") return nil
   end
   local sane= true
   for k, v in pairs(mats) do
      if type(k) ~= "string" then
         if type(k) == "number" then
            print("Material " .. k .. " has number key instead of name.")
            sane= false
         else
            print("Material has " .. type(k) " key instead of name.")
            sane= false
         end
      else
         if not sanity_check_single_material(k, v) then
            print("Material \"" .. k .. "\" failed sanity check.")
            sane= false
         end
      end
   end
   return sane
end

function reload_materials()
   local old_materials= materials
   dofile("materials.lua")
   if not sanity_check_materials(materials) then
      print("New materials not sane, reverting.")
      materials= old_materials
      old_materials= nil
   end
end

function write_materials()
   if not sanity_check_materials(materials) then print("Not attempting to write insane materials.") return nil end
   mfile,reason= io.open("materials.lua", "w")
   if not mfile then print(reason) return nil end
   mfile:write("-- Material format:\n")
   mfile:write("-- {string} name\n")
   mfile:write("-- {table} recipes\n")
   mfile:write("--   {table} recipe (unnamed array entries)\n")
   mfile:write("--     {string} station\n")
   mfile:write("--     {table} ingredient entry (unnamed array entries)\n")
   mfile:write("--       {number} amount\n")
   mfile:write("--       {string} name\n")
   mfile:write("--     ...\n")
   mfile:write("--   ...\n")
   mfile:write("-- {table} notes\n")
   mfile:write("--   {string} note\n")
   mfile:write("--   ...\n")
   mfile:write("materials= {\n")
   for k, v in pairs(materials) do
      mfile:write("  " .. k .. "= {name= \"" .. k .. "\", recipes= {")
      for n, rv in ipairs(v.recipes) do
         mfile:write("{station= \"" .. rv.station .. "\", ")
         for i, iv in ipairs(rv) do
            mfile:write("{amount= " .. iv.amount .. ", name= \"" .. iv.name .. "\"}")
         end
         mfile:write("}\n")
      end
      mfile:write("}},\n")
   end
   mfile:write("}\n")
end

reload_materials()
write_materials()
